"""Module for manage API"""
import json
import os
import urllib.request
import typing
from typing import Any, List, Optional, Type

from .annealing import AnnealingTask, AnnealingResult
from .data import ExecutionRequest, ExecutionRequestEncoder, TaskListData
from .device import Device
from .task import Task, TaskIter, TaskList

from .taskdatafactory import make_executiondata

if typing.TYPE_CHECKING:
    from blueqat import Circuit

API_ENDPOINT = "https://cloudapi.blueqat.com/"


class Api:
    """Manage API and post request."""
    def __init__(self, api_key: str):
        self.api_key = api_key

    def post_request(
            self,
            path: str,
            body: Any,
            json_encoder: Optional[Type[json.JSONEncoder]] = None) -> Any:
        """Post request."""
        headers = {
            'Content-Type': 'application/json',
            'X-Api-Key': self.api_key,
        }
        req = urllib.request.Request(
            API_ENDPOINT + path,
            json.dumps(body, cls=json_encoder).encode(), headers)
        with urllib.request.urlopen(req) as res:
            body = res.read()
        return json.loads(body)

    def save_api(self) -> None:
        """Save API to file."""
        d = os.path.join(os.environ["HOME"], ".bqcloud")
        os.makedirs(d, exist_ok=True)
        with open(os.path.join(d, "api_key"), "w") as f:
            f.write(self.api_key)

    def credit(self) -> str:
        """Get credit."""
        path = "v1/credit/get"
        return self.post_request(path, {})["amount"]

    def post_executiondata(self, execdata: ExecutionRequest) -> Task:
        """Create new task"""
        path = "v2/quantum-tasks/create"
        res = self.post_request(path, execdata, json_encoder=ExecutionRequestEncoder)
        return res

    def annealing(self, qubo: list[list[float]], chain_strength: int,
                  num_reads: int) -> AnnealingResult:
        """Create annealing task"""
        path = "v1/quantum-tasks/create"
        res = self.post_request(path, {
            "qubo": qubo,
            "chain_strength": chain_strength,
            "num_reads": num_reads
        })
        return AnnealingResult(**res)

    def execute(self,
                c: 'Circuit',
                device: Device,
                shots: int,
                group: Optional[str] = None,
                send_email: bool = False) -> Task:
        """Create new task and execute the task."""
        execdata = make_executiondata(c, device, shots, group, send_email)
        return self.post_executiondata(execdata)

    def annealing_tasks(self, index: int = 0) -> List[AnnealingTask]:
        """Get tasks."""
        path = "v1/quantum-tasks/list"
        body = {
            "index": index,
        }
        tasks = self.post_request(path, body)
        assert isinstance(tasks, list)
        return [AnnealingTask(self, **task) for task in tasks]

    def tasks(self, index: int = 0, per: Optional[int] = None) -> TaskList:
        """Get tasks."""
        path = "v2/quantum-tasks/list"
        body = {
            "index": index,
        }
        if per is not None and per > 0:
            body["per"] = per
        tasks = self.post_request(path, body)
        return TaskList(self, TaskListData(**tasks), index, per)

    def iter_tasks(self) -> TaskIter:
        """Get paginator of tasks"""
        return TaskIter(self)


def load_api() -> Api:
    """Load API from file."""
    with open(os.path.join(os.environ["HOME"], ".bqcloud/api_key")) as f:
        return Api(f.read().strip())


def register_api(api_key: str) -> Api:
    """Save and return API."""
    api = Api(api_key)
    api.save_api()
    return api



from typing import Optional

from blueqat import Circuit
from bqbraket import convert

from braket.device_schema import GateModelParameters
from braket.device_schema.ionq import IonqDeviceParameters
from braket.device_schema.rigetti import RigettiDeviceParameters
#from braket.device_schema.simulators import GateModelSimulatorDeviceParameters

from ..device import Device
from ..data import ExecutionRequest


def make_executiondata(c: Circuit, dev: Device, shots: int, group: Optional[str],
        send_email: bool) -> ExecutionRequest:
    action = convert(c).to_ir().json()
    dev_params = make_device_params(c, dev)
    return ExecutionRequest(action, dev.value, dev_params, shots, group, send_email)


def make_device_params(c: Circuit, dev: Device) -> str:
    paradigm_params = GateModelParameters(qubitCount=c.n_qubits,
                                          disableQubitRewiring=False)
    if "/rigetti/" in dev.value:
        return RigettiDeviceParameters(
            paradigmParameters=paradigm_params).json()
    if "/ionq/" in dev.value:
        return IonqDeviceParameters(paradigmParameters=paradigm_params).json()
    raise ValueError("Unknown AWS device.")



"""Module for task."""
import typing
from typing import Optional, SupportsIndex

from .data import TaskData, TaskListData
if typing.TYPE_CHECKING:
    from .api import Api


class Task:
    """Task."""
    def __init__(self, api: 'Api', taskdata: TaskData) -> None:
        self._api = api
        self.data = taskdata


class TaskList:
    """Task list."""
    def __init__(self, api: 'Api', tasklist: TaskListData, index: int, per: Optional[int]) -> None:
        self._api = api
        self.tasklist = tasklist
        self.index = index
        self.per = per

    @property
    def count(self) -> int:
        """Returns count of tasklist. It may larger than length of TaskList."""
        return self.tasklist.count

    def __getitem__(self, idx: SupportsIndex) -> Task:
        return Task(self._api, self.tasklist.tasks[idx])

    def __len__(self) -> int:
        return len(self.tasklist.tasks)

    def nextpage(self) -> 'TaskList':
        """Get next page of task list."""
        return self._api.tasks(self.index + 1, self.per)


class TaskIter:
    """Paginator of tasks."""
    def __init__(self, api: 'Api') -> None:
        self.tasklist = api.tasks()
        self.next_i = 0

    def __iter__(self) -> 'TaskIter':
        return self

    def __next__(self) -> Task:
        try:
            task = self.tasklist[self.next_i]
        except IndexError:
            self.tasklist = self.tasklist.nextpage()
            self.next_i = 0
            if len(self.tasklist) > 0:
                task = self.tasklist[self.next_i]
            else:
                raise StopIteration
        self.next_i += 1
        return task



"""Data classes and enums for communications."""
from dataclasses import dataclass
from enum import Enum
from json import JSONEncoder
from typing import Optional, SupportsIndex


@dataclass
class ExecutionRequest:
    """Internal data of task to be executed."""
    action: str
    device: str
    device_parameters: str
    shots: int
    task_group: Optional[str]
    send_email: bool


class ExecutionRequestEncoder(JSONEncoder):
    """JSONEncoder for ExecutionRequest"""
    def default(self, o):
        if isinstance(o, ExecutionRequest):
            return {
                'action': o.action,
                'device': o.device,
                'deviceParameters': o.device_parameters,
                'shots': o.shots,
                'taskGroup': o.task_group,
                'sendEmail': o.send_email
            }
        return super().default(o)


@dataclass
class TaskData:
    """Internal data of task object."""
    id: str
    userId: str
    action: str
    device: str
    deviceParameters: str
    shots: int
    createdAt: str
    updatedAt: str
    version: str


@dataclass
class TaskListData:
    """List of tasks which returns from API"""
    tasks: list[TaskData]
    count: int


class Status(str, Enum):
    """Statuses"""
    CREATED = "CREATED"
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLING = "CANCELLED"
    CANCELLED = "CANCELLED"

    def is_done(self) -> bool:
        """Check whether the status is finished (completed, failed or cancelled) or not"""
        return self in (Status.COMPLETED, Status.FAILED, Status.CANCELLED)



"""Module for making Task and TaskData.

This module is for internal use.
"""
import typing
from typing import Optional

from .device import Device
from .device_specific import aws_device
from .data import ExecutionRequest

if typing.TYPE_CHECKING:
    from blueqat import Circuit
    from .api import Api


def make_executiondata(c: 'Circuit',
                  dev: Device,
                  shots: int,
                  group: Optional[str] = None,
                  send_email: bool = False) -> ExecutionRequest:
    """Make ExecutionData for send job to the server."""
    if dev.value.startswith("aws/"):
        return aws_device.make_executiondata(c, dev, shots, group, send_email)
    raise ValueError(f"Cannot make {str(dev)} device task")



"""Module for annealing"""

from dataclasses import dataclass
from typing import Any


@dataclass
class AnnealingResult:
    """Result for annealing."""
    solutions: list[list[int]]
    solutionCounts: list[int]
    values: list[float]
    variableCount: int
    taskMetadata: dict[str, Any]
    additionalMetadata: dict[str, Any]


class AnnealingTask:
    """Annealing task."""
    def __init__(self, api, **kwargs) -> None:
        self.api = api
        self.data = kwargs

    def detail(self) -> AnnealingResult:
        """This method may be changed."""
        path = "v1/quantum-tasks/get"
        body = {
            "id": self.data['id'],
        }
        return AnnealingResult(**self.api.post_request(path, body))
